#ifndef STIM_BIP_H
#define STIM_BIP_H

#include "../envi/envi_header.h"
#include "../envi/bil.h"
#include "../envi/hsi.h"
#include <cstring>
#include <complex>
#include <utility>
#include <algorithm>

//CUDA
//#ifdef CUDA_FOUND
#include <stim/cuda/cudatools/error.h>
#include <cuda_runtime.h>
#include "cublas_v2.h"
#include "cufft.h"
//#endif

namespace stim{

/**
	The BIP class represents a 3-dimensional binary file stored using band interleaved by pixel (BIP) image encoding. The binary file is stored
	such that Z-X "frames" are stored sequentially to form an image stack along the y-axis. When accessing the data sequentially on disk,
	the dimensions read, from fastest to slowest, are Z, X, Y.

	This class is optimized for data streaming, and therefore supports extremely large (terabyte-scale) files. Data is loaded from disk
	on request. Functions used to access data are written to support efficient reading.
*/
template <typename T>

class bip: public hsi<T> {

protected:


	//std::vector<double> w; //band wavelength
	unsigned long long offset;		//header offset

	using hsi<T>::w;				//use the wavelength array in stim::hsi
	using hsi<T>::nnz;
	using binary<T>::progress;
	using hsi<T>::X;
	using hsi<T>::Y;
	using hsi<T>::Z;

public:

	using binary<T>::open;
	using binary<T>::file;
	using binary<T>::R;
	using binary<T>::read_line_0;

	bip(){ hsi<T>::init_bip(); }

	/// Open a data file for reading using the class interface.

	/// @param filename is the name of the binary file on disk
	/// @param X is the number of samples along dimension 1
	/// @param Y is the number of samples (lines) along dimension 2
	/// @param B is the number of samples (bands) along dimension 3
	/// @param header_offset is the number of bytes (if any) in the binary header
	/// @param wavelengths is an optional STL vector of size B specifying a numerical label for each band
	bool open(std::string filename,
			  unsigned long long X,
			  unsigned long long Y,
			  unsigned long long B,
			  unsigned long long header_offset,
			  std::vector<double> wavelengths,
			  stim::iotype io = stim::io_in){

		//copy the wavelengths to the BSQ file structure
		w = wavelengths;
		//copy the offset to the structure
		offset = header_offset;

		return open(filename, vec<unsigned long long>(B, X, Y), header_offset, io);

	}

	/// Retrieve a single band (based on index) and stores it in pre-allocated memory.

	/// @param p is a pointer to an allocated region of memory at least X * Y * sizeof(T) in size.
	/// @param page <= B is the integer number of the band to be copied.
	bool band_index( T * p, unsigned long long page, bool PROGRESS = false){
		return binary<T>::read_plane_0(p, page, PROGRESS);
	}

	/// Retrieve a single band (by numerical label) and stores it in pre-allocated memory.

	/// @param p is a pointer to an allocated region of memory at least X * Y * sizeof(T) in size.
	/// @param wavelength is a floating point value (usually a wavelength in spectral data) used as a label for the band to be copied.
	bool band( T * p, double wavelength, bool PROGRESS = false){

		//if there are no wavelengths in the BSQ file
		if(w.size() == 0)
			return band_index(p, (unsigned long long)wavelength, PROGRESS);

		unsigned long long XY = X() * Y();	//calculate the number of pixels in a band

		unsigned page=0;                      //bands around the wavelength


		//get the bands numbers around the wavelength

		//if wavelength is smaller than the first one in header file
		if ( w[page] > wavelength ){
			band_index(p, page, PROGRESS);
			return true;
		}

		while( w[page] < wavelength )
		{
			page++;
			//if wavelength is larger than the last wavelength in header file
			if (page == Z()) {
				band_index(p, Z()-1, PROGRESS);
				return true;
			}
		}
		if ( wavelength < w[page] )
		{
			T * p1;
			T * p2;
			p1=(T*)malloc( XY * sizeof(T));                     //memory allocation
			p2=(T*)malloc( XY * sizeof(T));
			band_index(p1, page - 1);
			band_index(p2, page, PROGRESS);
			for(unsigned long long i=0; i < XY; i++){
				double r = (double) (wavelength - w[page-1]) / (double) (w[page] - w[page-1]);
				p[i] = (T)(((double)p2[i] - (double)p1[i]) * r + (double)p1[i]);
			}
			free(p1);
			free(p2);
		}
		else                           //if the wavelength is equal to a wavelength in header file
		{
			band_index(p, page, PROGRESS);
		}
		return true;
	}

	/// Retrieve a single spectrum (Z-axis line) at a given (x, y) location and stores it in pre-allocated memory.

	/// @param p is a pointer to pre-allocated memory at least B * sizeof(T) in size.
	/// @param x is the x-coordinate (dimension 1) of the spectrum.
	/// @param y is the y-coordinate (dimension 2) of the spectrum.
	bool spectrum(T * p, unsigned long long x, unsigned long long y, bool PROGRESS = false){
		return read_line_0(p, x, y, PROGRESS);				//read a line in the binary YZ plane (dimension order for BIP is ZXY)
	}
	bool spectrum(T* p, size_t n, bool PROGRESS = false){
		size_t y = n / X();
		size_t x = n - y * X();
		return read_line_0(p, x, y, PROGRESS);				//read a line in the binary YZ plane (dimension order for BIP is ZXY)
	}

	/// Retrieves a band of x values from a given xz plane.

	/// @param p is a pointer to pre-allocated memory at least X * sizeof(T) in size
	/// @param c is a pointer to an existing XZ plane (size X*Z*sizeof(T))
	/// @param wavelength is the wavelength of X values to retrieve
	bool read_x_from_xz(T* p, T* c, double wavelength)
	{
		unsigned long long B = Z();

		unsigned long long page=0;                      //samples around the wavelength


		//get the bands numbers around the wavelength

		//if wavelength is smaller than the first one in header file
		if ( w[page] > wavelength ){
			for(unsigned long long j = 0; j < X(); j++)
			{
				p[j] = c[j * B];
			}
			return true;
		}

		while( w[page] < wavelength )
		{
			page++;
			//if wavelength is larger than the last wavelength in header file
			if (page == B) {
				for(unsigned long long j = 0; j < X(); j++)
				{
					p[j] = c[(j + 1) * B - 1];
				}
				return true;
			}
		}
		if ( wavelength < w[page] )
		{
			T * p1;
			T * p2;
			p1=(T*)malloc( X() * sizeof(T));                     //memory allocation
			p2=(T*)malloc( X() * sizeof(T));
			//band_index(p1, page - 1);
			for(unsigned long long j = 0; j < X(); j++)
			{
				p1[j] = c[j * B + page - 1];
			}
			//band_index(p2, page );
			for(unsigned long long j = 0; j < X(); j++)
			{
				p2[j] = c[j * B + page];
			}

			for(unsigned long long i=0; i < X(); i++){
				double r = (double) (wavelength - w[page-1]) / (double) (w[page] - w[page-1]);
				p[i] = (p2[i] - p1[i]) * r + p1[i];
			}
			free(p1);
			free(p2);
		}
		else                           //if the wavelength is equal to a wavelength in header file
		{
			//band_index(p, page);
			for(unsigned long long j = 0; j < X(); j++)
			{
				p[j] = c[j * B + page];
			}
		}

		return true;
	}

	/// Retrieve a single pixel and store it in a pre-allocated double array.
	bool pixeld(double* p, unsigned long long n){
		unsigned long long bandnum = X() * Y();		//calculate numbers in one band
		if ( n >= bandnum){							//make sure the pixel number is right
			std::cout<<"ERROR: sample or line out of range"<<std::endl;
			return false;
		}
		unsigned long long B = Z();

		T* temp = (T*) malloc(B * sizeof(T));						//allocate space for the raw pixel data
		file.seekg(n * B * sizeof(T), std::ios::beg);				//point to the certain pixel
		file.read((char *)temp, sizeof(T) * B);					//read the spectrum from disk to the temp pointer

		for(unsigned long long i = 0; i < B; i++)						//for each element of the spectrum
			p[i] = (double) temp[i];							//cast each element to a double value
		free(temp);												//free temporary memory
		return true;
	}

	/// Retrieve a single pixel and stores it in pre-allocated memory.

	/// @param p is a pointer to pre-allocated memory at least sizeof(T) in size.
	/// @param n is an integer index to the pixel using linear array indexing.
	bool pixel(T * p, unsigned long long n){

		unsigned long long N = X() * Y();					//calculate numbers in one band
		if ( n >= N){							//make sure the pixel number is right
			std::cout<<"ERROR: sample or line out of range"<<std::endl;
			return false;
		}

		file.seekg(n * Z() * sizeof(T), std::ios::beg);           //point to the certain pixel
		file.read((char *)p, sizeof(T) * Z());
		return true;
	}

	//given a Y ,return a ZX slice
	bool read_plane_y(T * p, size_t y){
		return binary<T>::read_plane_2(p, y);
	}

	/// Perform baseline correction given a list of baseline points and stores the result in a new BSQ file.

	/// @param outname is the name of the output file used to store the resulting baseline-corrected data.
	/// @param wls is the list of baseline points based on band labels.
	bool baseline(std::string outname, std::vector<double> base_pts, unsigned char* mask = NULL, bool PROGRESS = false){

		std::ofstream target(outname.c_str(), std::ios::binary);	//open the target binary file

		unsigned long long N = X() * Y();						//calculate the total number of pixels to be processed
		unsigned long long B = Z();								//get the number of bands
		T* s = (T*)malloc(sizeof(T) * B);						//allocate memory to store a pixel
		T* sbc = (T*)malloc(sizeof(T) * B);						//allocate memory to store the baseline corrected spectrum

		std::vector<T> base_vals;								//allocate space for the values at each baseline point
		double aw, bw;											//surrounding baseline point wavelengths
		T av, bv;												//surrounding baseline point values
		unsigned long long ai, bi;								//surrounding baseline point band indices
		for(unsigned long long n = 0; n < N; n++){				//for each pixel in the image
			if(mask != NULL && !mask[n]){						//if the pixel isn't masked
				memset(sbc, 0, sizeof(T) * B);					//set the baseline corrected spectrum to zero
			}
			else{

				pixel(s, n);										//retrieve the spectrum s
				base_vals = hsi<T>::interp_spectrum(s, base_pts);			//get the values at each baseline point

				ai = bi = 0;
				aw = w[0];											//initialize the current baseline points (assume the spectrum starts at 0)
				av = s[0];
				bw = base_pts[0];
				for(unsigned long long b = 0; b < B; b++){			//for each band in the spectrum
					while(bi < base_pts.size() && base_pts[bi] < w[b])	//while the current wavelength is higher than the second baseline point
						bi++;											//move to the next baseline point
					if(bi < base_pts.size()){
						bw = base_pts[bi];							//set the wavelength for the upper bound baseline point
						bv = base_vals[bi];							//set the value for the upper bound baseline point
					}
					if(bi == base_pts.size()){						//if we have passed the last baseline point
						bw = w[B-1];								//set the outer bound to the last spectral band
						bv = s[B-1];
					}
					if(bi != 0){
						ai = bi - 1;								//set the lower bound baseline point index
						aw = base_pts[ai];							//set the wavelength for the lower bound baseline point
						av = base_vals[ai];							//set the value for the lower bound baseline point
					}
					sbc[b] = s[b] - hsi<T>::lerp(w[b], av, aw, bv, bw);		//perform the baseline correction and save the new value
				}
			}

			if(PROGRESS) progress = (double)(n+1) / N * 100;	//set the current progress

			target.write((char*)sbc, sizeof(T) * B);	//write the corrected data into destination
		}														//end for each pixel

		free(s);
		free(sbc);
		target.close();

		return true;

	}

	/// Normalize all spectra based on the value of a single band, storing the result in a new BSQ file.

	/// @param outname is the name of the output file used to store the resulting baseline-corrected data.
	///	@param w is the label specifying the band that the hyperspectral image will be normalized to.
	///	@param t is a threshold specified such that a spectrum with a value at w less than t is set to zero. Setting this threshold allows the user to limit division by extremely small numbers.
	bool ratio(std::string outname, double w, unsigned char* mask = NULL, bool PROGRESS = false)
	{
		std::ofstream target(outname.c_str(), std::ios::binary);	//open the target binary file
		std::string headername = outname + ".hdr";              //the header file name

		unsigned long long N = X() * Y();						//calculate the total number of pixels to be processed
		unsigned long long B = Z();								//get the number of bands
		T* s = (T*)malloc(sizeof(T) * B);						//allocate memory to store a pixel
		T nv;													//stores the value of the normalized band
		for(unsigned long long n = 0; n < N; n++){				//for each pixel in the image
			if(mask != NULL && !mask[n])						//if the normalization band is below threshold
				memset(s, 0, sizeof(T) * B);					//set the output to zero
			else{
				pixel(s, n);										//retrieve the spectrum s
				nv = hsi<T>::interp_spectrum(s, w);							//find the value of the normalization band

				for(unsigned long long b = 0; b < B; b++)			//for each band in the spectrum
					s[b] /= nv;										//divide by the normalization value
			}

			if(PROGRESS) progress = (double)(n+1) / N * 100;	//set the current progress

			target.write((char*)s, sizeof(T) * B);	//write the corrected data into destination
		}														//end for each pixel

		free(s);
		target.close();
		return true;
	}

	void normalize(std::string outfile, unsigned char* mask = NULL, bool PROGRESS = false){

		std::ofstream target(outfile.c_str(), std::ios::binary);	//open the target binary file
		file.seekg(0, std::ios::beg);								//move the pointer to the current file to the beginning

		size_t B = Z();												//number of spectral components
		size_t XY = X() * Y();										//calculate the number of pixels
		size_t Bb = B * sizeof(T);									//number of bytes in a spectrum

		T* spec = (T*) malloc(Bb);									//allocate space for the spectrum
		T len;
		for(size_t xy = 0; xy < XY; xy++){							//for each pixel
			memset(spec, 0, Bb);									//set the spectrum to zero
			if(mask == NULL || mask[xy]){							//if the pixel is masked
				len = 0;											//initialize the
				file.read((char*)spec, Bb);							//read a spectrum
				for(size_t b = 0; b < B; b++)						//for each band
					len += spec[b]*spec[b];							//add the square of the spectral band
				len = sqrt(len);									//calculate the square of the sum of squared components
				for(size_t b = 0; b < B; b++)						//for each band
					spec[b] /= len;									//divide by the length
			}
			else
				file.seekg(Bb, std::ios::cur);						//otherwise skip a spectrum
			target.write((char*)spec, Bb);							//output the normalized spectrum
			if(PROGRESS) progress = (double)(xy + 1) / (double)XY * 100;		//update the progress
		}
	}

	/// This function loads a specified set of bands and saves them into a new output file
	bool select(std::string outfile, std::vector<double> bandlist, unsigned char* mask = NULL, bool PROGRESS = false) {
		std::ofstream target(outfile.c_str(), std::ios::binary);	//open the target binary file
		if (!target) {
			std::cout << "ERROR opening output file: " << outfile << std::endl;
			return false;
		}
		file.seekg(0, std::ios::beg);								//move the pointer to the current file to the beginning

		size_t B = Z();												//number of spectral components
		size_t XY = X() * Y();										//calculate the number of pixels
		size_t Bout = bandlist.size();
		size_t in_bytes = B * sizeof(T);							//number of bytes in a spectrum
		size_t out_bytes = Bout * sizeof(T);						//number of bytes in an output spectrum

		T* in = (T*)malloc(in_bytes);								//allocate space for the input spectrum
		T* out = (T*)malloc(out_bytes);								//allocate space for the output spectrum

		double wc;													//register to store the desired wavelength
		double w0, w1;												//registers to store the wavelengths surrounding the given band
		size_t b0, b1;												//indices of the bands surrounding the specified wavelength
		for (size_t xy = 0; xy < XY; xy++) {						//for each pixel
			//memset(out, 0, out_bytes);								//set the spectrum to zero
			if (mask == NULL || mask[xy]) {							//if the pixel is masked
				file.read((char*)in, in_bytes);						//read an input spectrum
				for (size_t b = 0; b < Bout; b++) {					//for each band
					wc = bandlist[b];									//set the desired wavelength
					hsi<T>::band_bounds(wc, b0, b1);								//get the surrounding bands
					w0 = w[b0];											//get the wavelength for the lower band
					w1 = w[b1];											//get the wavelength for the higher band
					out[b] = hsi<T>::lerp(wc, in[b0], w0, in[b1], w1);			//interpolate the spectral values to get the desired output value
				}
			}
			else
				file.seekg(Bout, std::ios::cur);						//otherwise skip a spectrum
			target.write((char*)out, out_bytes);							//output the normalized spectrum
			if (PROGRESS) progress = (double)(xy + 1) / (double)XY * 100;		//update the progress
		}

		free(in);
		free(out);
		return true;
	}


	/// Convert the current BIP file to a BIL file with the specified file name.

	/// @param outname is the name of the output BIL file to be saved to disk.
	bool bil(std::string outname, bool PROGRESS = false)
	{
		unsigned long long S = X() * Z() * sizeof(T);		//calculate the number of bytes in a ZX slice

		std::ofstream target(outname.c_str(), std::ios::binary);
		//std::string headername = outname + ".hdr";

		T * p;			//pointer to the current ZX slice for bip file
		p = (T*)malloc(S);
		T * q;			//pointer to the current XZ slice for bil file
		q = (T*)malloc(S);

		for ( unsigned long long i = 0; i < Y(); i++)
		{
			read_plane_y(p, i);
			for ( unsigned long long k = 0; k < Z(); k++)
			{
				unsigned long long ks = k * X();
				for ( unsigned long long j = 0; j < X(); j++)
					q[ks + j] = p[k + j * Z()];

				if(PROGRESS) progress = (double)(i * Z() + k+1) / (Y() * Z()) * 100;
			}
			target.write(reinterpret_cast<const char*>(q), S);   //write a band data into target file
		}

		free(p);
		free(q);
		target.close();
		return true;
	}

	/// Return a baseline corrected band given two adjacent baseline points and their bands. The result is stored in a pre-allocated array.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param lp is a pointer to an array holding the band image for the left baseline point
	/// @param rp is a pointer to an array holding the band image for the right baseline point
	/// @param wavelength is the label value for the requested baseline-corrected band
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size.
	bool baseline_band(double lb, double rb, T* lp, T* rp, double wavelength, T* result){

		unsigned long long XY = X() * Y();
		band(result, wavelength);		//get band

		//perform the baseline correction
		double r = (double) (wavelength - lb) / (double) (rb - lb);
		for(unsigned long long i=0; i < XY; i++){
			result[i] =(T) (result[i] - (rp[i] - lp[i]) * r - lp[i] );
		}
		return true;
	}

	/// Return a baseline corrected band given two adjacent baseline points. The result is stored in a pre-allocated array.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param bandwavelength is the label value for the desired baseline-corrected band
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size.
	bool height(double lb, double rb, double bandwavelength, T* result){

		T* lp;
		T* rp;
		unsigned long long XY = X() * Y();
		unsigned long long S = XY * sizeof(T);
		lp = (T*) malloc(S);			//memory allocation
		rp = (T*) malloc(S);

		band(lp, lb);
		band(rp, rb);

		baseline_band(lb, rb, lp, rp, bandwavelength, result);

		free(lp);
		free(rp);
		return true;
	}


	/// Calculate the area under the spectrum between two specified points and stores the result in a pre-allocated array.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param lab is the label value for the left bound (start of the integration)
	/// @param rab is the label value for the right bound (end of the integration)
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool area(double lb, double rb, double lab, double rab, T* result){

		T* lp;	//left band pointer
		T* rp;	//right band pointer
		T* cur;		//current band 1
		T* cur2;	//current band 2

		unsigned long long XY = X() * Y();
		unsigned long long S = XY * sizeof(T);

		lp = (T*) malloc(S);			//memory allocation
		rp = (T*) malloc(S);
		cur = (T*) malloc(S);
		cur2 = (T*) malloc(S);

		memset(result, (char)0, S);

		//find the wavelenght position in the whole band
		unsigned long long n = w.size();
		unsigned long long ai = 0;		//left bound position
		unsigned long long bi = n - 1;		//right bound position



		//to make sure the left and the right bound are in the bandwidth
		if (lb < w[0] || rb < w[0] || lb > w[n-1] || rb >w[n-1]){
			std::cout<<"ERROR: left bound or right bound out of bandwidth"<<std::endl;
			exit(1);
		}
		//to make sure rigth bound is bigger than left bound
		else if(lb > rb){
			std::cout<<"ERROR: right bound should be bigger than left bound"<<std::endl;
			exit(1);
		}

		//get the position of lb and rb
		while (lab >= w[ai]){
			ai++;
		}
		while (rab <= w[bi]){
			bi--;
		}

		band(lp, lb);
		band(rp, rb);

		//calculate the beginning and the ending part
		baseline_band(lb, rb, lp, rp, rab, cur2);		//ending part
		baseline_band(lb, rb, lp, rp, w[bi], cur);
		for(unsigned long long j = 0; j < XY; j++){
			result[j] += (T)((rab - w[bi]) * ((double)cur[j] + (double)cur2[j]) / 2.0);
		}
		baseline_band(lb, rb, lp, rp, lab, cur2);		//beginnning part
		baseline_band(lb, rb, lp, rp, w[ai], cur);
		for(unsigned long long j = 0; j < XY; j++){
			result[j] += (T)((w[ai] - lab) * ((double)cur[j] + (double)cur2[j]) / 2.0);
		}

		//calculate the area
		ai++;
		for(unsigned long long i = ai; i <= bi ;i++)
		{
			baseline_band(lb, rb, lp, rp, w[ai], cur2);
			for(unsigned long long j = 0; j < XY; j++)
			{
				result[j] += (T)((w[ai] - w[ai-1]) * ((double)cur[j] + (double)cur2[j]) / 2.0);
			}
			std::swap(cur,cur2);		//swap the band pointers
		}

		free(lp);
		free(rp);
		free(cur);
		free(cur2);
		return true;
	}

	/// Compute the ratio of two baseline-corrected peaks. The result is stored in a pre-allocated array.

	/// @param lb1 is the label value for the left baseline point for the first peak (numerator)
	/// @param rb1 is the label value for the right baseline point for the first peak (numerator)
	/// @param pos1 is the label value for the first peak (numerator) position
	/// @param lb2 is the label value for the left baseline point for the second peak (denominator)
	/// @param rb2 is the label value for the right baseline point for the second peak (denominator)
	/// @param pos2 is the label value for the second peak (denominator) position
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool ph_to_ph(T* result, double lb1, double rb1, double pos1, double lb2, double rb2, double pos2, unsigned char* mask = NULL){

		T* p1 = (T*)malloc(X() * Y() * sizeof(T));
		T* p2 = (T*)malloc(X() * Y() * sizeof(T));

		//get the two peak band
		height(lb1, rb1, pos1, p1);
		height(lb2, rb2, pos2, p2);
		//calculate the ratio in result
		for(unsigned long long i = 0; i < X() * Y(); i++){
			if(p1[i] == 0 && p2[i] ==0)
				result[i] = 1;
			else
				result[i] = p1[i] / p2[i];
		}

		free(p1);
		free(p2);
		return true;
	}

	/// Compute the ratio between a peak area and peak height.

	/// @param lb1 is the label value for the left baseline point for the first peak (numerator)
	/// @param rb1 is the label value for the right baseline point for the first peak (numerator)
	/// @param pos1 is the label value for the first peak (numerator) position
	/// @param lb2 is the label value for the left baseline point for the second peak (denominator)
	/// @param rb2 is the label value for the right baseline point for the second peak (denominator)
	/// @param pos2 is the label value for the second peak (denominator) position
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool pa_to_ph(T* result, double lb1, double rb1, double lab1, double rab1,
					double lb2, double rb2, double pos, unsigned char* mask = NULL){

		T* p1 = (T*)malloc(X() * Y() * sizeof(T));
		T* p2 = (T*)malloc(X() * Y() * sizeof(T));

		//get the area and the peak band
		area(lb1, rb1, lab1, rab1, p1);
		height(lb2, rb2, pos, p2);
		//calculate the ratio in result
		for(unsigned long long i = 0; i < X() * Y(); i++){
			if(p1[i] == 0 && p2[i] ==0)
				result[i] = 1;
			else
				result[i] = p1[i] / p2[i];
		}

		free(p1);
		free(p2);
		return true;
	}

	/// Compute the ratio between two peak areas.

	/// @param lb1 is the label value for the left baseline point for the first peak (numerator)
	/// @param rb1 is the label value for the right baseline point for the first peak (numerator)
	/// @param lab1 is the label value for the left bound (start of the integration) of the first peak (numerator)
	/// @param rab1 is the label value for the right bound (end of the integration) of the first peak (numerator)
	/// @param lb2 is the label value for the left baseline point for the second peak (denominator)
	/// @param rb2 is the label value for the right baseline point for the second peak (denominator)
	/// @param lab2 is the label value for the left bound (start of the integration) of the second peak (denominator)
	/// @param rab2 is the label value for the right bound (end of the integration) of the second peak (denominator)
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool pa_to_pa(T* result, double lb1, double rb1, double lab1, double rab1,
					double lb2, double rb2, double lab2, double rab2, unsigned char* mask = NULL){

		T* p1 = (T*)malloc(X() * Y() * sizeof(T));
		T* p2 = (T*)malloc(X() * Y() * sizeof(T));

		//get the area and the peak band
		area(lb1, rb1, lab1, rab1, p1);
		area(lb2, rb2, lab2, rab2, p2);
		//calculate the ratio in result
		for(unsigned long long i = 0; i < X() * Y(); i++){
			if(p1[i] == 0 && p2[i] ==0)
				result[i] = 1;
			else
				result[i] = p1[i] / p2[i];
		}

		free(p1);
		free(p2);
		return true;
	}

	/// Compute the definite integral of a baseline corrected peak.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param lab is the label for the start of the definite integral
	/// @param rab is the label for the end of the definite integral
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool x_area(double lb, double rb, double lab, double rab, T* result){
		T* lp;	//left band pointer
		T* rp;	//right band pointer
		T* cur;		//current band 1
		T* cur2;	//current band 2

		unsigned long long XY = X() * Y();
		unsigned long long S = XY * sizeof(T);

		lp = (T*) malloc(S);			//memory allocation
		rp = (T*) malloc(S);
		cur = (T*) malloc(S);
		cur2 = (T*) malloc(S);

		memset(result, (char)0, S);

		//find the wavelenght position in the whole band
		unsigned long long n = w.size();
		unsigned long long ai = 0;		//left bound position
		unsigned long long bi = n - 1;		//right bound position

		//to make sure the left and the right bound are in the bandwidth
		if (lb < w[0] || rb < w[0] || lb > w[n-1] || rb >w[n-1]){
			std::cout<<"ERROR: left bound or right bound out of bandwidth"<<std::endl;
			exit(1);
		}
		//to make sure rigth bound is bigger than left bound
		else if(lb > rb){
			std::cout<<"ERROR: right bound should be bigger than left bound"<<std::endl;
			exit(1);
		}

		//get the position of lb and rb
		while (lab >= w[ai]){
			ai++;
		}
		while (rab <= w[bi]){
			bi--;
		}

		band(lp, lb);
		band(rp, rb);

		//calculate the beginning and the ending part
		baseline_band(lb, rb, lp, rp, rab, cur2);		//ending part
		baseline_band(lb, rb, lp, rp, w[bi], cur);
		for(unsigned long long j = 0; j < XY; j++){
			result[j] += (T)((rab - w[bi]) * (rab + w[bi]) * ((double)cur[j] + (double)cur2[j]) / 4.0);
		}
		baseline_band(lb, rb, lp, rp, lab, cur2);		//beginnning part
		baseline_band(lb, rb, lp, rp, w[ai], cur);
		for(unsigned long long j = 0; j < XY; j++){
			result[j] += (T)((w[ai] - lab) * (w[ai] + lab) * ((double)cur[j] + (double)cur2[j]) / 4.0);
		}

		//calculate f(x) times x
		ai++;
		for(unsigned long long i = ai; i <= bi ;i++)
		{
			baseline_band(lb, rb, lp, rp, w[ai], cur2);
			for(unsigned long long j = 0; j < XY; j++)
			{
				result[j] += (T)((w[ai] - w[ai-1]) * (w[ai] + w[ai-1]) * ((double)cur[j] + (double)cur2[j]) / 4.0);
			}
			std::swap(cur,cur2);		//swap the band pointers
		}

		free(lp);
		free(rp);
		free(cur);
		free(cur2);
		return true;
	}

	/// Compute the centroid of a baseline corrected peak.

	/// @param lb is the label value for the left baseline point
	/// @param rb is the label value for the right baseline point
	/// @param lab is the label for the start of the peak
	/// @param rab is the label for the end of the peak
	/// @param result is a pointer to a pre-allocated array at least X * Y * sizeof(T) in size
	bool centroid(T* result, double lb, double rb, double lab, double rab, unsigned char* mask = NULL){
		T* p1 = (T*)malloc(X() * Y() * sizeof(T));
		T* p2 = (T*)malloc(X() * Y() * sizeof(T));

		//get the area and the peak band
		x_area(lb, rb, lab, rab, p1);
		area(lb, rb, lab, rab, p2);
		//calculate the ratio in result
		for(unsigned long long i = 0; i < X() * Y(); i++){
			if(mask == NULL || mask[i])
				result[i] = p1[i] / p2[i];
		}

		free(p1);
		free(p2);
		return true;
	}

	/// Create a mask based on a given band and threshold value.

	/// All pixels in the
	/// specified band greater than the threshold are true and all pixels less than the threshold are false.
	/// @param mask_band is the band used to specify the mask
	/// @param threshold is the threshold used to determine if the mask value is true or false
	/// @param p is a pointer to a pre-allocated array at least X * Y in size
	bool build_mask(unsigned char* out_mask, double mask_band, double lower, double upper, unsigned char* mask = NULL, bool PROGRESS = false){

		T* temp = (T*)malloc(X() * Y() * sizeof(T));		//allocate memory for the certain band
		band(temp, mask_band, PROGRESS);

		for (unsigned long long i = 0; i < X() * Y();i++) {
			if(mask == NULL || mask[i] != 0){
				if(temp[i] > lower && temp[i] < upper){
					out_mask[i] = 255;
				}
				else
					out_mask[i] = 0;
			}
		}

		free(temp);
		return true;

	}

	/// Apply a mask file to the BSQ image, setting all values outside the mask to zero.

	/// @param outfile is the name of the masked output file
	/// @param p is a pointer to memory of size X * Y, where p(i) = 0 for pixels that will be set to zero.
	bool apply_mask(std::string outfile, unsigned char* p, bool PROGRESS = false){

		std::ofstream target(outfile.c_str(), std::ios::binary);

		unsigned long long ZX = Z() * X();		//calculate the number of values in a page (XZ in BIP)
		unsigned long long L = ZX * sizeof(T);	//calculate the number of bytes in a page

		T * temp = (T*)malloc(L);		//allocate space for that page

		for (unsigned long long i = 0; i < Y(); i++)			//for each page (Y in BIP)
		{
			read_plane_y(temp, i);							//load that page (it's pointed to by temp)
			for ( unsigned long long j = 0; j < X(); j++)	//for each X value
			{
				for (unsigned long long k = 0; k < Z(); k++)	//for each B value (band)
				{
					if (p[i * X() + j] == 0)	//if the mask value is zero
					temp[j * Z() + k] = 0;			//set the pixel value to zero
				else								//otherwise just continue
					continue;
				}
			}
			target.write(reinterpret_cast<const char*>(temp), L);   //write the edited band data into target file
			if(PROGRESS) progress = (double)(i+1) / (double)Y() * 100;
		}
		target.close();						//close the target file
		free(temp);							//free allocated memory
		return true;						//return true
	}

	/// Copies all spectra corresponding to nonzero values of a mask into a pre-allocated matrix of size (B x P)
	///		where P is the number of masked pixels and B is the number of bands. The allocated memory can be accessed
	///		using the following indexing: i = p*B + b
	/// @param matrix is the destination for the pixel data
	/// @param mask is the mask
	bool sift(T* matrix, unsigned char* mask = NULL, bool PROGRESS = false){
		size_t Bbytes = sizeof(T) * Z();
		size_t XY = X() * Y();
		T* band = (T*) malloc( Bbytes );					//allocate space for a line

		file.seekg(0, std::ios::beg);	//seek to the beginning of the file

		size_t p = 0;										//create counter variables
		for(size_t xy = 0; xy < XY; xy++){					//for each pixel
			if(mask == NULL || mask[xy]){									//if the current pixel is masked
				file.read( (char*)band, Bbytes );			//read the current line
				for(size_t b = 0; b < Z(); b++){			//copy each band value to the sifted matrix
					size_t i = p * Z() + b;					//calculate the index in the sifted matrix
					matrix[i] = band[b];					//store the current value in the line at the correct matrix location
				}
				p++;									//increment the pixel pointer
			}
			else
				file.seekg(Bbytes, std::ios::cur);			//otherwise skip this band
			if(PROGRESS) progress = (double)(xy+1) / (double)XY * 100;
		}
		return true;
	}

	/// Saves to disk only those spectra corresponding to mask values != 0
	bool sift(std::string outfile, unsigned char* mask, bool PROGRESS = false){

		//reset the file pointer to the beginning of the file
		file.seekg(0, std::ios::beg);

		// open an output stream
		std::ofstream target(outfile.c_str(), std::ios::binary);

		//allocate space for a single spectrum
		unsigned long long B = Z();
		T* spectrum = (T*) malloc(B * sizeof(T));

		//calculate the number of pixels in a band
		unsigned long long XY = X() * Y();

		//for each pixel
		unsigned long long skip = 0;					//number of spectra to skip
		for(unsigned long long x = 0; x < XY; x++){

			//if the current pixel isn't masked
			if( mask[x] == 0){
				//increment the number of skipped pixels
				skip++;
			}
			//if the current pixel is masked
			else{

				//skip the intermediate pixels
				file.seekg(skip * B * sizeof(T), std::ios::cur);

				//set the skip value to zero
				skip = 0;

				//read this pixel into memory
				file.read((char *)spectrum, B * sizeof(T));

				//write this pixel out
				target.write((char *)spectrum, B * sizeof(T));
			}
			if(PROGRESS) progress = (double) (x+1) / XY * 100;

		}

		//close the output file
		target.close();
		free(spectrum);

		return true;
	}

	bool unsift(std::string outfile, unsigned char* mask, unsigned long long samples, unsigned long long lines, bool PROGRESS = false){

		// open an output stream
		std::ofstream target(outfile.c_str(), std::ios::binary);

		//reset the file pointer to the beginning of the file
		file.seekg(0, std::ios::beg);

		//allocate space for a single spectrum
		unsigned long long B = Z();
		T* spectrum = (T*) malloc(B * sizeof(T));

		//allocate space for a spectrum of zeros
		T* zeros = (T*) malloc(B * sizeof(T));
		memset(zeros, 0, B * sizeof(T));

		//calculate the number of pixels in a band
		unsigned long long XY = samples * lines;

		//for each pixel
		unsigned long long skip = 0;					//number of spectra to skip
		for(unsigned long long x = 0; x < XY; x++){

			//if the current pixel isn't masked
			if( mask[x] == 0){

				//write a bunch of zeros
				target.write((char *)zeros, B * sizeof(T));
			}
			//if the current pixel is masked
			else{

				//read a pixel into memory
				file.read((char *)spectrum, B * sizeof(T));

				//write this pixel out
				target.write((char *)spectrum, B * sizeof(T));
			}

			if(PROGRESS) progress = (double)(x + 1) / XY * 100;

		}

		//close the output file
		target.close();
		free(spectrum);

		//progress = 100;

		return true;


	}

	/// Calculate the mean value for all masked (or valid) pixels in a band and returns the average spectrum

	/// @param p is a pointer to pre-allocated memory of size [B * sizeof(T)] that stores the mean spectrum
	/// @param mask is a pointer to memory of size [X * Y] that stores the mask value at each pixel location
	bool mean_spectrum(double* m, double* std = NULL, unsigned char* mask = NULL, bool PROGRESS = false){
		unsigned long long XY = X() * Y();							//calculate the total number of pixels in the HSI
		T* temp = (T*)malloc(sizeof(T) * Z());						//allocate space for the current spectrum to be read
		memset(m, 0, Z() * sizeof(double));							//set the mean spectrum to zero
		double* e_x2 = (double*)malloc(Z() * sizeof(double));		//allocate space for E[x^2]
		memset(e_x2, 0, Z() * sizeof(double));						//set all values for E[x^2] to zero

		unsigned long long count = nnz(mask);									//calculate the number of masked pixels
		double x;
		for (unsigned long long i = 0; i < XY; i++){							//for each pixel in the HSI
			if (mask == NULL || mask[i] != 0){						//if the pixel is masked
				pixel(temp, i);										//get the spectrum
				for (unsigned long long j = 0; j < Z(); j++){					//for each spectral component
					x = temp[j];
					m[j] += x / (double)count;		//add the weighted value to the average
					e_x2[j] += x*x / (double)count;
				}
			}
			if(PROGRESS) progress = (double)(i+1) / XY * 100;		//increment the progress
		}

		//calculate the standard deviation
		if (std != NULL) {
			for (size_t i = 0; i < Z(); i++)
				std[i] = sqrt(e_x2[i] - m[i] * m[i]);
		}

		free(temp);
		return true;
	}
//#ifdef CUDA_FOUND
	/// Calculate the covariance matrix for masked pixels using cuBLAS
	/// Note that cuBLAS only supports integer-sized arrays, so there may be issues with large spectra
	int co_matrix_cublas(double* co, double* avg, unsigned char *mask, bool PROGRESS = false){

		cudaError_t cudaStat;
		cublasStatus_t stat;
		cublasHandle_t handle;

		unsigned long long XY = X() * Y();									//calculate the number of elements in a band image
		unsigned long long B = Z();											//calculate the number of spectral elements

		double* s = (double*)malloc(sizeof(double) * B);					//allocate space for the spectrum that will be pulled from the file
		double* s_dev;														//declare a device pointer that will store the spectrum on the GPU
		double* A_dev;														//declare a device pointer that will store the covariance matrix on the GPU
		double* avg_dev;													//declare a device pointer that will store the average spectrum
		cudaStat = cudaMalloc(&s_dev, B * sizeof(double));					//allocate space on the CUDA device for the spectrum
		cudaStat = cudaMalloc(&A_dev, B * B * sizeof(double));				//allocate space on the CUDA device for the covariance matrix
		cudaStat = cudaMemset(A_dev, 0, B * B * sizeof(double));			//initialize the covariance matrix to zero (0)
		cudaStat = cudaMalloc(&avg_dev, B * sizeof(double));				//allocate space on the CUDA device for the average spectrum
		stat = cublasSetVector((int)B, sizeof(double), avg, 1, avg_dev, 1);		//copy the average spectrum to the CUDA device

		double ger_alpha = 1.0/(double)XY;									//scale the outer product by the inverse of the number of samples (mean outer product)
		double axpy_alpha = -1;												//multiplication factor for the average spectrum (in order to perform a subtraction)

		stat = cublasCreate(&handle);										//create a cuBLAS instance
		if (stat != CUBLAS_STATUS_SUCCESS) return 1;						//test the cuBLAS instance to make sure it is valid

		//else std::cout<<"Using cuBLAS to calculate the mean covariance matrix..."<<std::endl;
		for (unsigned long long xy = 0; xy < XY; xy++){										//for each pixel
			if (mask == NULL || mask[xy] != 0){
				pixeld(s, xy);																	//retreive the spectrum at the current xy pixel location
				stat = cublasSetVector((int)B, sizeof(double), s, 1, s_dev, 1);						//copy the spectrum from the host to the device
				stat = cublasDaxpy(handle, (int)B, &axpy_alpha, avg_dev, 1, s_dev, 1);				//subtract the average spectrum
				stat = cublasDsyr(handle, CUBLAS_FILL_MODE_UPPER, (int)B, &ger_alpha, s_dev, 1, A_dev, (int)B);	//calculate the covariance matrix (symmetric outer product)
			}
			if(PROGRESS) progress = (double)(xy+1) / XY * 100;													//record the current progress

		}

		cublasGetMatrix((int)B, (int)B, sizeof(double), A_dev, (int)B, co, (int)B);					//copy the result from the GPU to the CPU

		cudaFree(A_dev);														//clean up allocated device memory
		cudaFree(s_dev);
		cudaFree(avg_dev);

		for(unsigned long long i = 0; i < B; i++){										//copy the upper triangular portion to the lower triangular portion
			for(unsigned long long j = i+1; j < B; j++){
				co[B * i + j] = co[B * j + i];
			}
		}

		return 0;
	}
//#endif

	/// Calculate the covariance matrix for all masked pixels in the image with 64-bit floating point precision.

	/// @param co is a pointer to pre-allocated memory of size [B * B] that stores the resulting covariance matrix
	/// @param avg is a pointer to memory of size B that stores the average spectrum
	/// @param mask is a pointer to memory of size [X * Y] that stores the mask value at each pixel location
	bool co_matrix(double* co, double* avg, unsigned char *mask, int cuda_device = 0, bool PROGRESS = false){
		progress = 0;

		if(cuda_device >= 0){													//if a CUDA device is specified
			int dev_count;
			HANDLE_ERROR(cudaGetDeviceCount(&dev_count));						//get the number of CUDA devices
			if(dev_count > 0 && dev_count > cuda_device){							//if the first device is not an emulator
				cudaDeviceProp prop;
				cudaGetDeviceProperties(&prop, cuda_device);									//get the property of the requested CUDA device
				if (prop.major != 9999) {
					std::cout << "Using CUDA device [" << cuda_device << "] to calculate the mean covariance matrix..."<<std::endl;
					HANDLE_ERROR(cudaSetDevice(cuda_device));
					int status = co_matrix_cublas(co, avg, mask, PROGRESS);			//use cuBLAS to calculate the covariance matrix
					if (status == 0) return true;									//if the cuBLAS function returned correctly, we're done
				}
			}																	//otherwise continue using the CPU		
			std::cout<<"WARNING: cuBLAS failed, using CPU"<<std::endl;
		}
		//memory allocation
		unsigned long long XY = X() * Y();
		unsigned long long B = Z();
		T* temp = (T*)malloc(sizeof(T) * B);

		unsigned long long count = nnz(mask);								//count the number of masked pixels

		//initialize covariance matrix
		memset(co, 0, B * B * sizeof(double));

		//calculate covariance matrix
		double* co_half = (double*) malloc(B * B * sizeof(double));			//allocate space for a higher-precision intermediate matrix
		double* temp_precise = (double*) malloc(B * sizeof(double));
		memset(co_half, 0, B * B * sizeof(double));							//initialize the high-precision matrix with zeros
		unsigned long long idx;													//stores i*B to speed indexing
		for (unsigned long long xy = 0; xy < XY; xy++){
			if (mask == NULL || mask[xy] != 0){
				pixel(temp, xy);												//retreive the spectrum at the current xy pixel location
				for(unsigned long long b = 0; b < B; b++)									//subtract the mean from this spectrum and increase the precision
					temp_precise[b] = (double)temp[b] - (double)avg[b];
				idx = 0;
				for (unsigned long long b0 = 0; b0 < B; b0++){								//for each band
					for (unsigned long long b1 = b0; b1 < B; b1++)
						co_half[idx++] += temp_precise[b0] * temp_precise[b1];
				}
			}
			if(PROGRESS) progress = (double)(xy+1) / XY * 100;
		}
		idx = 0;
		for (unsigned long long i = 0; i < B; i++){										//copy the precision matrix to both halves of the output matrix
			for (unsigned long long j = i; j < B; j++){
				co[j * B + i] = co[i * B + j] = co_half[idx++] / (double) count;
			}
		}

		free(temp);
		free(temp_precise);
		return true;
	}


	int coNoise_matrix_cublas(double* coN, double* avg, unsigned char *mask, bool PROGRESS = false) {

		cudaError_t cudaStat;
		cublasStatus_t stat;
		cublasHandle_t handle;

		progress = 0;													    //initialize the progress to zero (0)
		unsigned long long XY = X() * Y();									//calculate the number of elements in a band image
		unsigned long long B = Z();											//calculate the number of spectral elements

		double* s = (double*)malloc(sizeof(double) * B);					//allocate space for the spectrum that will be pulled from the file
		double* s_dev;														//declare a device pointer that will store the spectrum on the GPU

		double* s2 = (double*)malloc(sizeof(double) * B);					//allocate space for the spectrum of second pixel that will be pulled from the file
		double* s2_dev;														//  device pointer on the GPU
		cudaStat = cudaMalloc(&s2_dev, B * sizeof(double));					//  allocate space on the CUDA device
		cudaStat = cudaMemset(s2_dev, 0, B * sizeof(double));               //  initialize s2_dev to zero (0)

		double* A_dev;														//declare a device pointer that will store the covariance matrix on the GPU
		double* avg_dev;													//declare a device pointer that will store the average spectrum
		cudaStat = cudaMalloc(&s_dev, B * sizeof(double));					//allocate space on the CUDA device for the spectrum
		cudaStat = cudaMalloc(&A_dev, B * B * sizeof(double));				//allocate space on the CUDA device for the covariance matrix
		cudaStat = cudaMemset(A_dev, 0, B * B * sizeof(double));			//initialize the covariance matrix to zero (0)
		cudaStat = cudaMalloc(&avg_dev, B * sizeof(double));				//allocate space on the CUDA device for the average spectrum
		stat = cublasSetVector((int)B, sizeof(double), avg, 1, avg_dev, 1);		//copy the average spectrum to the CUDA device

		double ger_alpha = 1.0 / (double)XY;									//scale the outer product by the inverse of the number of samples (mean outer product)
		double axpy_alpha = -1;												//multiplication factor for the average spectrum (in order to perform a subtraction)

		CUBLAS_HANDLE_ERROR(cublasCreate(&handle));							//create a cuBLAS instance
		if (stat != CUBLAS_STATUS_SUCCESS) return 1;						//test the cuBLAS instance to make sure it is valid

		for (unsigned long long xy = 0; xy < XY; xy++) {										//for each pixel
			if (mask == NULL || mask[xy] != 0) {
				pixeld(s, xy);                                        //retreive the spectrum at the current xy pixel location
				if (xy < XY - X()) {
					pixeld(s2, xy + X());                              //retreive the spectrum at the current xy+X pixel location, which is adjacent (bellow) to the pixel at xy location (in y direction)
				}
				else {
					pixeld(s2, xy - X());                               //for the last row we consider the the adjacent pixel which is located above pixel xy
				}
				stat = cublasSetVector((int)B, sizeof(double), s, 1, s_dev, 1);						//copy the spectrum of first pixel from the host to the device
				stat = cublasDaxpy(handle, (int)B, &axpy_alpha, avg_dev, 1, s_dev, 1);				//subtract the average spectrum

				stat = cublasSetVector((int)B, sizeof(double), s2, 1, s2_dev, 1);					//copy the spectrum of second pixel from the host to the device
				stat = cublasDaxpy(handle, (int)B, &axpy_alpha, avg_dev, 1, s2_dev, 1);				//subtract the average spectrum

				stat = cublasDaxpy(handle, (int)B, &axpy_alpha, s2_dev, 1, s_dev, 1);	   //Minimum/Maximum Autocorrelation Factors (MAF) method : subtranct each pixel from adjacent pixel (in y direction)

				stat = cublasDsyr(handle, CUBLAS_FILL_MODE_UPPER, (int)B, &ger_alpha, s_dev, 1, A_dev, (int)B);	//calculate the covariance matrix (symmetric outer product)
			}
			if (PROGRESS) progress = (double)(xy + 1) / XY * 100;													//record the current progress

		}

		cublasGetMatrix((int)B, (int)B, sizeof(double), A_dev, (int)B, coN, (int)B);					//copy the result from the GPU to the CPU

		cudaFree(A_dev);														//clean up allocated device memory
		cudaFree(s_dev);
		cudaFree(s2_dev);
		cudaFree(avg_dev);

		for (unsigned long long i = 0; i < B; i++) {										//copy the upper triangular portion to the lower triangular portion
			for (unsigned long long j = i + 1; j < B; j++) {
				coN[B * i + j] = coN[B * j + i];
			}
		}

		return 0;
	}
	//#endif

	/// Calculate the covariance of noise matrix for all masked pixels in the image with 64-bit floating point precision.

	/// @param coN is a pointer to pre-allocated memory of size [B * B] that stores the resulting covariance matrix
	/// @param avg is a pointer to memory of size B that stores the average spectrum
	/// @param mask is a pointer to memory of size [X * Y] that stores the mask value at each pixel location
	bool coNoise_matrix(double* coN, double* avg, unsigned char *mask, int cuda_device = 0, bool PROGRESS = false) {

		if (cuda_device >= 0) {													//if a CUDA device is specified
			int dev_count;
			HANDLE_ERROR(cudaGetDeviceCount(&dev_count));						//get the number of CUDA devices
			if (dev_count > 0 && dev_count > cuda_device) {							//if the first device is not an emulator
				cudaDeviceProp prop;
				cudaGetDeviceProperties(&prop, cuda_device);									//get the property of the requested CUDA device
				if (prop.major != 9999) {
					std::cout << "Using CUDA device [" << cuda_device << "] to calculate the noise covariance matrix..." << std::endl;
					HANDLE_ERROR(cudaSetDevice(cuda_device));
					int status = coNoise_matrix_cublas(coN, avg, mask, PROGRESS);			//use cuBLAS to calculate the covariance matrix
					if (status == 0) return true;									//if the cuBLAS function returned correctly, we're done
				}
			}																	//otherwise continue using the CPU
			std::cout << "WARNING: cuBLAS failed, using CPU" << std::endl;
		}

		progress = 0;
		//memory allocation
		unsigned long long XY = X() * Y();
		unsigned long long B = Z();
		T* temp = (T*)malloc(sizeof(T) * B);
		T* temp2 = (T*)malloc(sizeof(T) * B);

		unsigned long long count = nnz(mask);								//count the number of masked pixels

																			//initialize covariance matrix of noise
		memset(coN, 0, B * B * sizeof(double));

		//calculate covariance matrix
		double* coN_half = (double*)malloc(B * B * sizeof(double));			//allocate space for a higher-precision intermediate matrix
		double* temp_precise = (double*)malloc(B * sizeof(double));
		double* temp_precise2 = (double*)malloc(B * sizeof(double));
		memset(coN_half, 0, B * B * sizeof(double));							//initialize the high-precision matrix with zeros
		unsigned long long idx;													//stores i*B to speed indexing
		for (unsigned long long xy = 0; xy < XY; xy++) {
			if (mask == NULL || mask[xy] != 0) {
				pixel(temp, xy);                                        //retreive the spectrum at the current xy pixel location
				if (xy < XY - X()) {
					pixel(temp2, xy + X());                              //retreive the spectrum at the current xy+X pixel location, which is adjacent (bellow) to the pixel at xy location (in y direction)
				}
				else {
					pixel(temp2, xy - X());                               //for the last row we consider the the adjacent pixel which is located above pixel xy
				}
				for (unsigned long long b = 0; b < B; b++) {									//subtract the mean from this spectrum and increase the precision
					temp_precise[b] = (double)temp[b] - (double)avg[b];
					temp_precise2[b] = (double)temp2[b] - (double)avg[b];
				}

				for (unsigned long long b2 = 0; b2 < B; b2++)	    //Minimum/Maximum Autocorrelation Factors (MAF) method : subtranct each pixel from adjacent pixel (in y direction)
					temp_precise[b2] -= temp_precise2[b2];

				idx = 0;
				for (unsigned long long b0 = 0; b0 < B; b0++) {								//for each band
					for (unsigned long long b1 = b0; b1 < B; b1++)
						coN_half[idx++] += temp_precise[b0] * temp_precise[b1];
				}
			}
			if (PROGRESS) progress = (double)(xy + 1) / XY * 100;
		}
		idx = 0;
		for (unsigned long long i = 0; i < B; i++) {										//copy the precision matrix to both halves of the output matrix
			for (unsigned long long j = i; j < B; j++) {
				coN[j * B + i] = coN[i * B + j] = coN_half[idx++] / (double)count;
			}
		}

		free(temp);
		free(temp2);
		free(temp_precise);
		free(temp_precise2);
		return true;
	}

    /// Project the spectra onto a set of basis functions
	/// @param outfile is the name of the new binary output file that will be created
	/// @param center is a spectrum about which the data set will be rotated (ex. when performing mean centering)
	/// @param basis a set of basis vectors that the data set will be projected onto (after centering)
	/// @param M is the number of basis vectors
	/// @param mask is a character mask used to limit processing to valid pixels
	bool project_cublas(std::string outfile, double* center, double* basis, unsigned long long M, unsigned char* mask = NULL, bool PROGRESS = false){

		//cudaError_t cudaStat;
		//cublasStatus_t stat;
		cublasHandle_t handle;

		std::ofstream target(outfile.c_str(), std::ios::binary);	//open the target binary file

		progress = 0;													    //initialize the progress to zero (0)
		unsigned long long XY = X() * Y();									//calculate the number of elements in a band image
		unsigned long long B = Z();											//calculate the number of spectral elements

		double* s = (double*)malloc(sizeof(double) * B);					//allocate space for the spectrum that will be pulled from the file
		double* s_dev;														//declare a device pointer that will store the spectrum on the GPU
		HANDLE_ERROR(cudaMalloc(&s_dev, B * sizeof(double)));				//allocate space on the CUDA device for the spectrum


        double* basis_dev;														//  device pointer on the GPU
        HANDLE_ERROR(cudaMalloc(&basis_dev, M * B * sizeof(double)));					//  allocate space on the CUDA device
        HANDLE_ERROR(cudaMemset(basis_dev, 0, M * B * sizeof(double)));               //  initialize basis_dev to zero (0)


        /// transposing basis matrix (because cuBLAS is column-major)
        double *basis_Transposed = (double*)malloc(M * B * sizeof(double));
        memset(basis_Transposed, 0, M * B * sizeof(double));
        for (int i = 0; i<M; i++)
            for (int j = 0; j<B; j++)
            basis_Transposed[i+j*M] = basis[i*B+j];
		//copy the basis_Transposed matrix to the CUDA device (both matrices are stored in column-major format)
		CUBLAS_HANDLE_ERROR(cublasSetMatrix((int)M, (int)B, sizeof(double),basis_Transposed, (int)M, basis_dev, (int)M));

		double* center_dev;																						//declare a device pointer that will store the center (average)
		HANDLE_ERROR(cudaMalloc(&center_dev, B * sizeof(double)));									//allocate space on the CUDA device for the center (average)
		CUBLAS_HANDLE_ERROR(cublasSetVector((int)B, sizeof(double), center, 1, center_dev, 1));			//copy the center vector (average) to the CUDA device (from host to device)


        double* A = (double*)malloc(sizeof(double) * M);								//allocate space for the projected pixel on the host
        double* A_dev;																	//declare a device pointer that will store the projected pixel on the GPU
		HANDLE_ERROR(cudaMalloc(&A_dev,M * sizeof(double)));				    //allocate space on the CUDA device for the projected pixel
		HANDLE_ERROR(cudaMemset(A_dev, 0,M * sizeof(double)));		        //initialize the projected pixel to zero (0)

		double axpy_alpha = -1;												//multiplication factor for the center (in order to perform a subtraction)
		double axpy_alpha2 = 1;												//multiplication factor for the matrix-vector multiplication
        double axpy_beta = 0;												//multiplication factor for the matrix-vector multiplication (there is no second scalor)

		CUBLAS_HANDLE_ERROR(cublasCreate(&handle));					//create a cuBLAS instance

        T* temp = (T*)malloc(sizeof(T) * M);													//allocate space for the projected pixel to be written on the disc
		size_t i;
		for (unsigned long long xy = 0; xy < XY; xy++){											//for each pixel
			if (mask == NULL || mask[xy] != 0){
				pixeld(s, xy);																	//retreive the spectrum at the current xy pixel location

				CUBLAS_HANDLE_ERROR(cublasSetVector((int)B, sizeof(double), s, 1, s_dev, 1));						    //copy the spectrum from the host to the device
                CUBLAS_HANDLE_ERROR(cublasDaxpy(handle, (int)B, &axpy_alpha, center_dev, 1, s_dev, 1));					//subtract the center (average)
                CUBLAS_HANDLE_ERROR(cublasDgemv(handle,CUBLAS_OP_N,(int)M,(int)B,&axpy_alpha2,basis_dev,(int)M,s_dev,1,&axpy_beta,A_dev,1));         //performs the matrix-vector multiplication
                CUBLAS_HANDLE_ERROR(cublasGetVector((int)M, sizeof(double), A_dev, 1, A, 1));							//copy the projected pixel to the host (from GPU to CPU)
						
				for(i = 0; i < M; i++)	temp[i] = (T)A[i];										//casting projected pixel from double to whatever T is
			}
			else
				memset(temp, 0, sizeof(T) * M);

			target.write(reinterpret_cast<const char*>(temp), sizeof(T) * M);					  //write the projected vector
			if(PROGRESS) progress = (double)(xy+1) / XY * 100;									    //record the current progress

		}

        //clean up allocated device memory
		HANDLE_ERROR(cudaFree(A_dev));
		HANDLE_ERROR(cudaFree(s_dev));
		HANDLE_ERROR(cudaFree(basis_dev));
		HANDLE_ERROR(cudaFree(center_dev));
		CUBLAS_HANDLE_ERROR(cublasDestroy(handle));
		free(A);
		free(s);
		free(temp);
		target.close();												//close the output file
		
		return true;
	}

	/// Project the spectra onto a set of basis functions
	/// @param outfile is the name of the new binary output file that will be created
	/// @param center is a spectrum about which the data set will be rotated (ex. when performing mean centering)
	/// @param basis a set of basis vectors that the data set will be projected onto (after centering)
	/// @param M is the number of basis vectors
	/// @param mask is a character mask used to limit processing to valid pixels
	bool project(std::string outfile, double* center, double* basis, unsigned long long M, unsigned char* mask = NULL, int cuda_device = 0, bool PROGRESS = false){
		if (cuda_device >= 0) {													//if a CUDA device is specified
			int dev_count;
			HANDLE_ERROR(cudaGetDeviceCount(&dev_count));						//get the number of CUDA devices
			if (dev_count > 0 && dev_count > cuda_device) {							//if the first device is not an emulator
				cudaDeviceProp prop;
				cudaGetDeviceProperties(&prop, cuda_device);									//get the property of the requested CUDA device
				if (prop.major != 9999) {
					std::cout << "Using CUDA device [" << cuda_device << "] to perform a basis projection..." << std::endl;
					HANDLE_ERROR(cudaSetDevice(cuda_device));
					return project_cublas(outfile, center, basis, M, mask, PROGRESS);
				}
			}																	//otherwise continue using the CPU		
			std::cout << "WARNING: cuBLAS failed, using CPU" << std::endl;
		}
		
		std::ofstream target(outfile.c_str(), std::ios::binary);	//open the target binary file
		//std::string headername = outfile + ".hdr";					//the header file name

		//memory allocation
		unsigned long long XY = X() * Y();
		unsigned long long B = Z();

		T* s = (T*)malloc(sizeof(T) * B);							//allocate space for the spectrum
		T* rs = (T*)malloc(sizeof(T) * M);							//allocate space for the projected spectrum
		double* bv;													//pointer to the current projection vector
		for(unsigned long long xy = 0; xy < XY; xy++){				//for each spectrum in the image
			memset(rs, 0, sizeof(T) * M);
			if(mask == NULL || mask[xy]){
				pixel(s, xy);											//load the spectrum
				for(unsigned long long m = 0; m < M; m++){				//for each basis vector
					bv = &basis[m * B];									//assign 'bv' to the beginning of the basis vector
					for(unsigned long long b = 0; b < B; b++){			//for each band
						rs[m] += (T)(((double)s[b] - center[b]) * bv[b]);		//center the spectrum and perform the projection
					}
				}
			}

			target.write(reinterpret_cast<const char*>(rs), sizeof(T) * M);					//write the projected vector
			if(PROGRESS) progress = (double)(xy+1) / XY * 100;
		}

		free(s);													//free temporary storage arrays
		free(rs);
		target.close();												//close the output file
		
		return true;
	}

	bool inverse(std::string outfile, double* center, double* basis, unsigned long long B, unsigned long long C = 0, bool PROGRESS = false){

		std::ofstream target(outfile.c_str(), std::ios::binary);	//open the target binary file
		std::string headername = outfile + ".hdr";					//the header file name

		//memory allocation
		unsigned long long XY = X() * Y();
		if(C == 0) C = Z();											//if no coefficient number is given, assume all are used
		C = std::min<unsigned long long>(C, Z());					//set the number of coefficients (the user can specify fewer)

		T* coeff = (T*)malloc(sizeof(T) * Z());						//allocate space for the coefficients
		T* s = (T*)malloc(sizeof(T) * B);							//allocate space for the spectrum
		double* bv;													//pointer to the current projection vector
		for(unsigned long long xy = 0; xy < XY; xy++){				//for each pixel in the image (expressed as a set of coefficients)
			pixel(coeff, xy);										//load the coefficients
			memset(s, 0, sizeof(T) * B);							//initialize the spectrum to zero (0)
			for(unsigned long long c = 0; c < C; c++){				//for each basis vector coefficient
				bv = &basis[c * B];									//assign 'bv' to the beginning of the basis vector
				for(unsigned long long b = 0; b < B; b++){			//for each component of the basis vector
					s[b] += (T)((double)coeff[c] * bv[b] + center[b]);			//calculate the contribution of each element of the basis vector in the final spectrum
				}
			}

			target.write(reinterpret_cast<const char*>(s), sizeof(T) * B);					//write the projected vector
			if(PROGRESS) progress = (double)(xy+1) / XY * 100;
		}

		free(coeff);												//free temporary storage arrays
		free(s);
		target.close();												//close the output file

		return true;
	}


	/// Crop a region of the image and save it to a new file.

	/// @param outfile is the file name for the new cropped image
	/// @param x0 is the lower-left x pixel coordinate to be included in the cropped image
	/// @param y0 is the lower-left y pixel coordinate to be included in the cropped image
	/// @param x1 is the upper-right x pixel coordinate to be included in the cropped image
	/// @param y1 is the upper-right y pixel coordinate to be included in the cropped image
	bool crop(std::string outfile, unsigned long long x0,
								   unsigned long long y0,
								   unsigned long long x1,
								   unsigned long long y1,
								   unsigned long long b0,
								   unsigned long long b1,
								   bool PROGRESS = false){

		//calculate the new number of samples, lines, and bands
		unsigned long long samples = x1 - x0 + 1;
		unsigned long long lines = y1 - y0 + 1;
		unsigned long long bands = b1 - b0 + 1;

		//calculate the length of one cropped spectrum
		unsigned long long L = bands * sizeof(T);

		//unsigned long long L = Z() * sizeof(T);

		//allocate space for the spectrum
		char* temp = (char*)malloc(L);

		//open an output file for binary writing
		std::ofstream out(outfile.c_str(), std::ios::binary);

		//seek to the first pixel in the cropped image
		size_t startx = x0 * Z();
		size_t starty = y0 * X() * Z();
		size_t startb = b0;
		file.seekg( (starty + startx + startb) * sizeof(T), std::ios::beg);

		//distance between sample spectra in the same line
		size_t dist_between_samples = Z() - bands;
		size_t jump_sample = dist_between_samples * sizeof(T);

		//distance between sample spectra in adjacent lines
		//unsigned long long jump_line = ( X() - x1 + x0 ) * Z() * sizeof(T);
		size_t dist_between_lines = X() - samples;
		size_t jump_line = dist_between_lines * Z() * sizeof(T);


		//unsigned long long sp = y0 * X() + x0;		//start pixel

		//for each pixel in the image
		for (unsigned y = 0; y < lines; y++) {
			for (unsigned x = 0; x < samples; x++) {
				//read the cropped spectral region
				file.read(temp, L );
				//pixel(temp, sp + x + y * X());
				out.write(temp, L);   //write slice data into target file

				file.seekg(jump_sample, std::ios::cur);

				if(PROGRESS) progress = (double)(y * samples + x + 1) / (lines * samples) * 100;
			}

			file.seekg(jump_line, std::ios::cur);
		}
		free(temp);
		out.close();

		return true;
	}

	/// Remove a list of bands from the ENVI file

	/// @param outfile is the file name for the output hyperspectral image (with trimmed bands)
	/// @param b is an array of bands to be eliminated
	void trim(std::string outfile, std::vector<size_t> band_array, bool PROGRESS = false){

		std::ofstream out(outfile.c_str(), std::ios::binary);	//open the output file for writing
		file.seekg(0, std::ios::beg);							//move to the beginning of the input file

		size_t B = Z();								//calculate the number of elements in a spectrum
		size_t Bdst = Z() - band_array.size();		//calculate the number of elements in an output spectrum
		size_t Bb = B * sizeof(T);					//calculate the number of bytes in a spectrum
		size_t XY = X() * Y();						//calculate the number of pixels in the image
		T* src = (T*)malloc(Bb);					//allocate space to store an input spectrum
		T* dst = (T*)malloc(Bdst * sizeof(T));		//allocate space to store an output spectrum

		size_t i;									//index into the band array
		size_t bdst;								//index into the output array
		for(size_t xy = 0; xy < XY; xy++){			//for each pixel
			i = 0;
			bdst = 0;
			file.read((char*)src, Bb);				//read a spectrum
			for(size_t b = 0; b < B; b++){			//for each band
				if(b != band_array[i]){				//if the band isn't trimmed
					dst[bdst] = src[b];				//copy the band value to the output spectrum
					bdst++;
				}
				else i++;							//otherwise increment i
			}
			out.write((char*)dst, Bdst * sizeof(T));	//write the output spectrum
			if(PROGRESS) progress = (double)(xy + 1) / (double) XY * 100;
		}
		free(src);
		free(dst);
	}

	/// Combine two BIP images along the Y axis

	/// @param outfile is the combined file to be output
	/// @param infile is the input file stream for the image to combine with this one
	/// @param Sx is the size of the second image along X
	/// @param Sy is the size of the second image along Y
	/// @param offset is a shift (negative or positive) in the combined image to the left or right
	void combine(std::string outfile, bip<T>* C, long long xp, long long yp, bool PROGRESS = false){
		std::ofstream out(outfile.c_str(), std::ios::binary);	//open the output file for writing
		file.seekg(0, std::ios::beg);								//move to the beginning of both files
		C->file.seekg(0, std::ios::beg);

		size_t S[2];				//size of the output band image
		size_t p0[2];				//position of the current image in the output
		size_t p1[2];				//position of the source image in the output

		hsi<T>::calc_combined_size(xp, yp, C->X(), C->Y(), S[0], S[1], p0[0], p0[1], p1[0], p1[1]);	//calculate the image placement parameters

		size_t spec_bytes = Z() * sizeof(T);						//calculate the number of bytes in a spectrum
		T* spec = (T*)malloc(spec_bytes);							//allocate space for a spectrum

		for(size_t y = 0; y < S[1]; y++){							//for each pixel in the destination image
			for(size_t x = 0; x < S[0]; x++){
				if(x >= p0[0] && x < p0[0] + X() && y >= p0[1] && y < p0[1] + Y())	//if this pixel is in the current image
					file.read((char*)spec, spec_bytes);
				else if(x >= p1[0] && x < p1[0] + C->X() && y >= p1[1] && y < p1[1] + C->Y())	//if this pixel is in the source image
					C->file.read((char*)spec, spec_bytes);
				else
					memset(spec, 0, spec_bytes);
				out.write((char*)spec, spec_bytes);					//write to the output file
			}
			if(PROGRESS) progress = (double)( (y+1) * S[0] + 1) / (double) (S[0] * S[1]) * 100;
		}
	}

	///Append two files together along the band dimension
	void append(std::string outfile, bip<T>* C, bool PROGRESS = false) {
		std::ofstream out(outfile.c_str(), std::ios::binary);	//open the output file for writing
		file.seekg(0, std::ios::beg);							//move to the beginning of both files
		C->file.seekg(0, std::ios::beg);
		size_t a_bytes = Z() * sizeof(T);					//calculate the number of bytes in a single plane of this file
		size_t b_bytes = C->Z() * sizeof(T);			//calculate the number of bytes in a single plane of the appending file
		T* a = (T*)malloc(a_bytes);								//allocate space for a plane of the current file
		T* b = (T*)malloc(b_bytes);								//allocate space for a plane of the appended file
		if (PROGRESS) progress = 0;
		for (size_t xy = 0; xy < X()*Y(); xy++) {
			spectrum(a, xy);								//read a plane from the current file
			out.write((char*)a, a_bytes);								//write the plane to disk
			C->spectrum(b, xy);								//read a plane from the appending file
			out.write((char*)b, b_bytes);
			if (PROGRESS) progress = (double)(xy + 1) / (double)(X() * Y()) * 100;
		}

		out.close();
	}
	/// Convolve the given band range with a kernel specified by a vector of coefficients.

	/// @param outfile is an already open stream to the output file
	/// @param C is an array of coefficients
	/// @param start is the band to start processing (the first coefficient starts here)
	/// @param nbands is the number of bands to process
	/// @param center is the index for the center coefficient for the kernel (used to set the wavelengths in the output file)

	void convolve(std::string outfile, std::vector<double> C, size_t start, size_t end, unsigned char* mask = NULL, bool PROGRESS = false){
		std::ofstream out(outfile.c_str(), std::ios::binary);		//open the output file for writing

		size_t N = end - start + 1;									//number of bands in the output spectrum
		size_t Nb = N * sizeof(T);									//size of the output spectrum in bytes
		size_t B = Z();												//calculate the number of values in a spectrum
		size_t Bb = B * sizeof(T);									//calculate the size of a spectrum in bytes

		file.seekg(0, std::ios::beg);								//move to the beginning of the input file

		size_t nC = C.size();										//get the number of bands that the kernel spans
		T* inspec = (T*)malloc(Bb);									//allocate space for the input spectrum
		T* outspec = (T*)malloc(Nb);								//allocate space for the output spectrum

		size_t XY = X() * Y();										//number of pixels in the image
		for(size_t xy = 0; xy < XY; xy++){							//for each pixel
			file.read((char*)inspec, Bb);							//read an input spectrum
			memset(outspec, 0, Nb);									//set the output spectrum to zero (0)
			if(mask == NULL || mask[xy]){
				for(size_t b = 0; b < N; b++){							//for each component of the spectrum
					for(size_t c = 0; c < nC; c++){						//for each coefficient in the kernel
						outspec[b] += (T)(inspec[b + start + c] * C[c]);		//perform the sum/multiply part of the convolution
					}
				}
			}
			out.write((char*)outspec, Nb);							//output the band
			if(PROGRESS) progress = (double)(xy+1) / (double)XY * 100;
		}
	}

	void deriv(std::string outfile, size_t d, size_t order, const std::vector<double> w, unsigned char* mask = NULL, bool PROGRESS = false){
		std::ofstream out(outfile.c_str(), std::ios::binary);		//open the output file for writing


		size_t B = Z();												//calculate the number of values in a spectrum
		size_t Bb = B * sizeof(T);									//calculate the size of a spectrum in bytes

		bool UNIFORM = true;
		double ds = w[1] - w[0];									//initialize ds
		for(size_t b = 1; b < B; b++)								//test to see if the spectral spacing is uniform (if it is, convolution is much faster)
			if(w[b] - w[b-1] != ds) UNIFORM = false;

		size_t nC = order + d;									//approximating a derivative requires order + d samples

		file.seekg(0, std::ios::beg);								//move to the beginning of the input file

		T* inspec = (T*)malloc(Bb);									//allocate space for the input spectrum
		T* outspec = (T*)malloc(Bb);								//allocate space for the output spectrum

		size_t XY = X() * Y();										//number of pixels in the image
		size_t mid = (size_t)(nC / 2);							//calculate the mid point of the kernel
		size_t iw;													//index to the first wavelength used to evaluate the derivative at this band
		for(size_t xy = 0; xy < XY; xy++){							//for each pixel
			file.read((char*)inspec, Bb);							//read an input spectrum
			memset(outspec, 0, Bb);									//set the output spectrum to zero (0)
			if(mask == NULL || mask[xy]){
				iw = 0;
				for(size_t b = 0; b < mid; b++){							//for each component of the spectrum
					std::vector<double> w_pts(w.begin() + iw, w.begin() + iw + nC);			//get the wavelengths corresponding to each sample
					std::vector<double> C = diff_coefficients(w[b], w_pts, d);					//get the optimal sample weights
					for(size_t c = 0; c < nC; c++)						//for each coefficient in the kernel
						outspec[b] += (T)(inspec[iw + c] * C[c]);		//perform the sum/multiply part of the convolution
				}
				std::vector<double> w_pts(w.begin(), w.begin() + nC);			//get the wavelengths corresponding to each sample
				std::vector<double> C = diff_coefficients(w[0], w_pts, d);					//get the optimal sample weights
				for(size_t b = mid; b <= B - (nC - mid); b++){
					iw = b - mid;
					if(!UNIFORM){																//if the spacing is non-uniform, we have to re-calculate these points every iteration
						std::vector<double> w_pts(w.begin() + iw, w.begin() + iw + nC);			//get the wavelengths corresponding to each sample
						std::vector<double> C = diff_coefficients(w[b], w_pts, d);					//get the optimal sample weights
					}
					for(size_t c = 0; c < nC; c++)						//for each coefficient in the kernel
						outspec[b] += (T)(inspec[iw + c] * C[c]);		//perform the sum/multiply part of the convolution
				}
				iw = B - nC;
				for(size_t b = B - (nC - mid) + 1; b < B; b++){
					std::vector<double> w_pts(w.begin() + iw, w.begin() + iw + nC);			//get the wavelengths corresponding to each sample
					std::vector<double> C = diff_coefficients(w[b], w_pts, d);					//get the optimal sample weights
					for(size_t c = 0; c < nC; c++)						//for each coefficient in the kernel
						outspec[b] += (T)(inspec[iw + c] * C[c]);		//perform the sum/multiply part of the convolution
				}
			}
			out.write((char*)outspec, Bb);							//output the band
			if(PROGRESS) progress = (double)(xy+1) / (double)XY * 100;
		}
	}

	bool multiply(std::string outname, double v, unsigned char* mask = NULL, bool PROGRESS = false){
		std::ofstream target(outname.c_str(), std::ios::binary);		//open the target binary file
		std::string headername = outname + ".hdr";						//the header file name

		unsigned long long N = X() * Y();								//calculate the total number of pixels to be processed
		unsigned long long B = Z();										//get the number of bands
		T* s = (T*)malloc(sizeof(T) * B);								//allocate memory to store a pixel
		for(unsigned long long n = 0; n < N; n++){						//for each pixel in the image
			if(mask == NULL || mask[n]){								//if the pixel is masked
				for(size_t b = 0; b < B; b++)							//for each band in the spectrum
					s[b] *= (T)v;											//multiply
			}

			if(PROGRESS) progress = (double)(n+1) / N * 100;			//set the current progress

			target.write((char*)s, sizeof(T) * B);						//write the corrected data into destination
		}																//end for each pixel

		free(s);														//free the spectrum
		target.close();													//close the output file
		return true;
	}

	bool add(std::string outname, double v, unsigned char* mask = NULL, bool PROGRESS = false){
		std::ofstream target(outname.c_str(), std::ios::binary);		//open the target binary file
		std::string headername = outname + ".hdr";						//the header file name

		unsigned long long N = X() * Y();								//calculate the total number of pixels to be processed
		unsigned long long B = Z();										//get the number of bands
		T* s = (T*)malloc(sizeof(T) * B);								//allocate memory to store a pixel
		for(unsigned long long n = 0; n < N; n++){						//for each pixel in the image
			if(mask == NULL || mask[n]){								//if the pixel is masked
				for(size_t b = 0; b < B; b++)							//for each band in the spectrum
					s[b] += (T)v;											//multiply
			}

			if(PROGRESS) progress = (double)(n+1) / N * 100;			//set the current progress

			target.write((char*)s, sizeof(T) * B);						//write the corrected data into destination
		}																//end for each pixel

		free(s);														//free the spectrum
		target.close();													//close the output file
		return true;
	}

	int fft(std::string outname, size_t bandmin, size_t bandmax, size_t samples = 0, T* ratio = NULL, size_t rx = 0, size_t ry = 0, bool PROGRESS = false, int device = 0){
		if(device == -1){
			std::cout<<"ERROR: GPU required for FFT (uses cuFFT)."<<std::endl;
			exit(1);
		}
		if(samples == 0) samples = Z();								//if samples are specified, use all of them
		if(samples > Z()){
			std::cout<<"ERROR: stim::envi doesn't support FFT padding just yet."<<std::endl;
			exit(1);
		}
		int nd;															//stores the number of CUDA devices
		HANDLE_ERROR(cudaGetDeviceCount(&nd));							//get the number of CUDA devices
		if(device >= nd){												//test for the existence of the requested device
			std::cout<<"ERROR: requested CUDA device for stim::envi::fft() doesn't exist"<<std::endl;
			exit(1);
		}
		HANDLE_ERROR(cudaSetDevice(device));							//set the CUDA device
		cudaDeviceProp prop;
		HANDLE_ERROR(cudaGetDeviceProperties(&prop, device));			//get the CUDA device properties

		size_t B = Z();
		size_t S = samples;
		size_t fft_size = S * sizeof(T);								//number of bytes for each FFT
		size_t cuda_bytes = prop.totalGlobalMem;						//get the number of bytes of global memory available
		size_t cuda_use = (size_t)floor(cuda_bytes * 0.2);								//only use 80%
		size_t nS = cuda_use / fft_size;								//calculate the number of spectra that can be loaded onto the GPU as a single batch
		size_t batch_bytes = nS * fft_size;								//calculate the size of a batch (in bytes)
		size_t fft_bytes = nS * (S/2 + 1) * sizeof(cufftComplex);
		T* batch = (T*) malloc(batch_bytes);							//allocate space in host memory to store a batch
		memset(batch, 0, batch_bytes);
		std::complex<T>* batch_fft = (std::complex<T>*) malloc(fft_bytes);
		T* gpu_batch;													//device pointer to the batch
		HANDLE_ERROR(cudaMalloc(&gpu_batch, batch_bytes));				//allocate space on the device for the FFT batch
		cufftComplex* gpu_batch_fft;												//allocate space for the FFT result
		HANDLE_ERROR(cudaMalloc(&gpu_batch_fft, fft_bytes));
		int N[1];														//create an array with the interferogram size (required for cuFFT input)
		N[0] = (int)S;													//set the only array value to the interferogram size

		//if a background is provided for a ratio
		std::complex<T>* ratio_fft = NULL;											//create a pointer for the FFT of the ratio image (if it exists)
		if(ratio){
			size_t bkg_bytes = rx * ry * S * sizeof(T);								//calculate the total number of bytes in the background image
			T* bkg_copy = (T*) malloc(bkg_bytes);									//allocate space to copy the background
			if(S == Z()) memcpy(bkg_copy, ratio, bkg_bytes);						//if the number of samples used in processing equals the number of available samples
			else{
				for(size_t xyi = 0; xyi < rx*ry; xyi++)
					memcpy(&bkg_copy[xyi * S], &ratio[xyi * B], S * sizeof(T));
			}
			T* gpu_ratio;
			HANDLE_ERROR(cudaMalloc(&gpu_ratio, bkg_bytes));
			HANDLE_ERROR(cudaMemcpy(gpu_ratio, bkg_copy, bkg_bytes, cudaMemcpyHostToDevice));
			cufftHandle bkg_plan;
			CUFFT_HANDLE_ERROR(cufftPlanMany(&bkg_plan, 1, N, NULL, 1, N[0], NULL, 1, N[0], CUFFT_R2C, (int)(rx * ry)));
			size_t bkg_fft_bytes = rx * ry * (S / 2 + 1) * sizeof(cufftComplex);
			T* gpu_ratio_fft;
			HANDLE_ERROR(cudaMalloc(&gpu_ratio_fft, bkg_fft_bytes));
			CUFFT_HANDLE_ERROR(cufftExecR2C(bkg_plan, (cufftReal*)gpu_ratio, (cufftComplex*)gpu_ratio_fft));
			ratio_fft = (std::complex<T>*) malloc(bkg_fft_bytes);
			HANDLE_ERROR(cudaMemcpy(ratio_fft, gpu_ratio_fft, bkg_fft_bytes, cudaMemcpyDeviceToHost));
			HANDLE_ERROR(cudaFree(gpu_ratio));
			HANDLE_ERROR(cudaFree(gpu_ratio_fft));
			CUFFT_HANDLE_ERROR(cufftDestroy(bkg_plan));
		}

		cufftHandle plan;												//create a CUFFT plan
		CUFFT_HANDLE_ERROR(cufftPlanMany(&plan, 1, N, NULL, 1, N[0], NULL, 1, N[0], CUFFT_R2C, (int)nS));

		std::ofstream outfile(outname, std::ios::binary);				//open a file for writing

		size_t XY = X() * Y();											//calculate the number of spectra
		size_t xy = 0;
		size_t bs;														//stores the number of spectra in the current batch
		size_t s, b;
		size_t S_fft = S/2 + 1;
		size_t bandkeep = bandmax - bandmin + 1;
		size_t x, y;
		size_t ratio_i;
		T* temp_spec = (T*) malloc(Z() * sizeof(T));					//allocate space to hold a single pixel
		while(xy < XY){													//while there are unprocessed spectra
			bs = (std::min)(XY - xy, nS);										//calculate the number of spectra to include in the batch
			for(s = 0; s < bs; s++){									//for each spectrum in the batch
				pixel(temp_spec, xy + s);						//read a pixel from disk
				memcpy(&batch[s * S], temp_spec, S * sizeof(T));
				//pixel(&batch[s * S], xy + s);							//read the next spectrum
			}
			HANDLE_ERROR(cudaMemcpy(gpu_batch, batch, batch_bytes, cudaMemcpyHostToDevice));
			CUFFT_HANDLE_ERROR(cufftExecR2C(plan, (cufftReal*)gpu_batch, gpu_batch_fft));			//execute the (implicitly forward) transform
			HANDLE_ERROR(cudaMemcpy(batch_fft, gpu_batch_fft, fft_bytes, cudaMemcpyDeviceToHost));	//copy the data back to the GPU
			for(s = 0; s < bs; s++){															//for each spectrum in the batch
				y = (xy + s)/X();
				x = xy + s - y * X();
				if(ratio_fft)	ratio_i = (y % ry) * rx + (x % rx);								//if a background is used, calculate the coordinates into it
				for(b = 0; b < S/2 + 1; b++){														//for each sample
					if(ratio_fft)						
						batch[s * S + b] = -log(abs(batch_fft[s * S_fft + b]) / abs(ratio_fft[ratio_i * S_fft + b]));
					else
						batch[s * S + b] = abs(batch_fft[s * S_fft + b]);		//calculate the magnitude of the spectrum					
				}
				outfile.write((char*)&batch[s * S + bandmin], bandkeep * sizeof(T));							//save the resulting spectrum
			}
			xy += bs;													//increment xy by the number of spectra processed
			if(PROGRESS) progress = (double)xy / (double)XY * 100;
		}
		outfile.close();
		free(ratio_fft);
		free(batch_fft);
		free(batch);
		HANDLE_ERROR(cudaFree(gpu_batch));
		HANDLE_ERROR(cudaFree(gpu_batch_fft));
		return 0;
	}

	/// Close the file.
	bool close(){
		file.close();
		return true;
	}

	};
}

#endif
